@@ -575,7 +575,7 @@ grn_ctx_impl_init(grn_ctx *ctx)
   ctx->impl->finalizer = NULL;
 
   ctx->impl->com = NULL;
-  ctx->impl->outbuf = grn_obj_open(ctx, GRN_BULK, 0, 0);
+  ctx->impl->outbuf = grn_obj_open(ctx, GRN_BULK, 0, GRN_DB_TEXT);
   ctx->impl->output = NULL;
   ctx->impl->data.ptr = NULL;
   ctx->impl->tv.tv_sec = 0;
@@ -1397,6 +1397,7 @@ grn_ctx_qe_exec_uri(grn_ctx *ctx, const char *path, uint32_t path_len)
     grn_expr_exec(ctx, expr, 0);
   }
 exit :
+  GRN_OBJ_FIN(ctx, &request_id);
   GRN_OBJ_FIN(ctx, &buf);
   return expr;
 }
@@ -2398,9 +2399,8 @@ grn_calloc_default(grn_ctx *ctx, size_t size, const char* file, int line, const
       grn_alloc_info_add(res, file, line, func);
     } else {
       if (!(res = calloc(size, 1))) {
-        MERR("calloc fail (%" GRN_FMT_LLU ")=%p (%s:%d) <%" GRN_FMT_LLU ">",
-             (unsigned long long int)size, res, file, line,
-             (unsigned long long int)alloc_count);
+        MERR("calloc fail (%" GRN_FMT_SIZE ")=%p (%s:%d) <%d>",
+             size, res, file, line, alloc_count);
       } else {
         GRN_ADD_ALLOC_COUNT(1);
         grn_alloc_info_add(res, file, line, func);
