@@ -24,6 +24,14 @@
 
 #include <string.h>
 #include <stdio.h>
+#include <stdlib.h>
+#include <fcntl.h>
+#include <sys/stat.h>
+
+#ifdef WIN32
+# include <io.h>
+# include <share.h>
+#endif /* WIN32 */
 
 grn_rc
 grn_normalize_offset_and_limit(grn_ctx *ctx, int size, int *p_offset, int *p_limit)
@@ -212,8 +220,13 @@ grn_inspect_type(grn_ctx *ctx, grn_obj *buf, unsigned char type)
     break;
   default:
     {
-      char type_in_hex[5]; /* "0xXX" */
-      sprintf(type_in_hex, "%#02x", type);
+#define TYPE_IN_HEX_SIZE 5 /* "0xXX" */
+      char type_in_hex[TYPE_IN_HEX_SIZE];
+      grn_snprintf(type_in_hex,
+                   TYPE_IN_HEX_SIZE,
+                   TYPE_IN_HEX_SIZE,
+                   "%#02x", type);
+#undef TYPE_IN_HEX_SIZE
       GRN_TEXT_PUTS(ctx, buf, "(unknown: ");
       GRN_TEXT_PUTS(ctx, buf, type_in_hex);
       GRN_TEXT_PUTS(ctx, buf, ")");
@@ -1295,7 +1308,7 @@ grn_win32_base_dir(void)
                                                     absolute_dll_filename,
                                                     MAX_PATH);
     if (absolute_dll_filename_size == 0) {
-      win32_base_dir = strdup(".");
+      win32_base_dir = grn_strdup_raw(".");
     } else {
       DWORD ansi_dll_filename_size;
       ansi_dll_filename_size =
@@ -1303,7 +1316,7 @@ grn_win32_base_dir(void)
                             absolute_dll_filename, absolute_dll_filename_size,
                             NULL, 0, NULL, NULL);
       if (ansi_dll_filename_size == 0) {
-        win32_base_dir = strdup(".");
+        win32_base_dir = grn_strdup_raw(".");
       } else {
         char *path;
         win32_base_dir = malloc(ansi_dll_filename_size + 1);
@@ -1316,8 +1329,8 @@ grn_win32_base_dir(void)
           *path = '\0';
         }
         path = strrchr(win32_base_dir, '\\');
-        if (path && (strcasecmp(path + 1, "bin") == 0 ||
-                     strcasecmp(path + 1, "lib") == 0)) {
+        if (path && (grn_strcasecmp(path + 1, "bin") == 0 ||
+                     grn_strcasecmp(path + 1, "lib") == 0)) {
           *path = '\0';
         } else {
           path = win32_base_dir + strlen(win32_base_dir);
@@ -1334,3 +1347,72 @@ grn_win32_base_dir(void)
   return win32_base_dir;
 }
 #endif
+
+#ifdef WIN32
+int
+grn_mkstemp(char *path_template)
+{
+  errno_t error;
+  size_t path_template_size;
+  int fd;
+
+  path_template_size = strlen(path_template) + 1;
+  error = _mktemp_s(path_template, path_template_size);
+  if (error != 0) {
+    return -1;
+  }
+
+  error = _sopen_s(&fd,
+                   path_template,
+                   _O_RDWR | _O_CREAT | _O_EXCL | _O_BINARY,
+                   _SH_DENYNO,
+                   _S_IREAD | _S_IWRITE);
+  if (error != 0) {
+    return -1;
+  }
+
+  return fd;
+}
+#else /* WIN32 */
+int
+grn_mkstemp(char *path_template)
+{
+# ifdef HAVE_MKSTEMP
+  return mkstemp(path_template);
+# else /* HAVE_MKSTEMP */
+  mktemp(path_template);
+  return open(path_template, O_RDWR | O_CREAT | O_EXCL, S_IRUSR | S_IWUSR);
+# endif /* HAVE_MKSTEMP */
+}
+#endif /* WIN32 */
+
+#ifdef WIN32
+UINT
+grn_windows_encoding_to_code_page(grn_encoding encoding)
+{
+  UINT code_page;
+
+  switch (encoding) {
+  case GRN_ENC_EUC_JP :
+    code_page = 20932;
+    break;
+  case GRN_ENC_UTF8 :
+    code_page = CP_UTF8;
+    break;
+  case GRN_ENC_SJIS :
+    code_page = 932;
+    break;
+  case GRN_ENC_LATIN1 :
+    code_page = 1252;
+    break;
+  case GRN_ENC_KOI8R :
+    code_page = 20866;
+    break;
+  default :
+    code_page = CP_ACP;
+    break;
+  }
+
+  return code_page;
+}
+#endif /* WIN32 */
