@@ -39,9 +39,9 @@
 #ifdef HAVE_SYS_SOCKET_H
 # include <sys/socket.h>
 #endif /* HAVE_SYS_SOCKET_H */
-#ifdef HAVE_NETINET_IN_H
+#ifndef WIN32
 # include <netinet/in.h>
-#endif /* HAVE_NETINET_IN_H */
+#endif /* WIN32 */
 
 #ifdef HAVE_SYS_RESOURCE_H
 # include <sys/resource.h>
@@ -53,17 +53,11 @@
 
 #ifdef WIN32
 # include <io.h>
+# include <direct.h>
 #else /* WIN32 */
 # include <sys/uio.h>
 #endif /* WIN32 */
 
-#ifdef HAVE__STRNICMP
-# ifdef strncasecmp
-#  undef strncasecmp
-# endif /* strcasecmp */
-# define strncasecmp(s1,s2,n) _strnicmp(s1,s2,n)
-#endif /* HAVE__STRNICMP */
-
 #ifndef USE_MSG_NOSIGNAL
 # ifdef MSG_NOSIGNAL
 #  undef MSG_NOSIGNAL
@@ -109,7 +103,7 @@ static int ready_notify_pipe[2];
 static grn_encoding encoding;
 static grn_command_version default_command_version;
 static int64_t default_match_escalation_threshold;
-static int log_level;
+static grn_bool use_windows_event_log = GRN_FALSE;
 
 static int
 grn_rc_to_exit_code(grn_rc rc)
@@ -363,6 +357,9 @@ do_alone(int argc, char **argv)
   grn_obj *db;
   grn_ctx ctx_, *ctx = &ctx_;
   grn_ctx_init(ctx, 0);
+  if (use_windows_event_log) {
+    grn_windows_event_logger_set(ctx);
+  }
   if (argc > 0 && argv) { path = *argv++; argc--; }
   db = (newdb || !path) ? grn_db_create(ctx, path, NULL) : grn_db_open(ctx, path);
   if (db) {
@@ -440,6 +437,9 @@ g_client(int argc, char **argv)
   const char *hostname = DEFAULT_DEST;
   if (argc > 0 && argv) { hostname = *argv++; argc--; }
   grn_ctx_init(ctx, 0);
+  if (use_windows_event_log) {
+    grn_windows_event_logger_set(ctx);
+  }
   if (!grn_ctx_connect(ctx, hostname, port, 0)) {
     if (!argc) {
       grn_obj text;
@@ -477,6 +477,32 @@ static grn_mutex q_mutex;
 static grn_cond q_cond;
 static uint32_t nthreads = 0, nfthreads = 0, max_nfthreads;
 
+static uint32_t
+groonga_get_thread_count(void *data)
+{
+  return max_nfthreads;
+}
+
+static void
+groonga_set_thread_count(uint32_t new_count, void *data)
+{
+  uint32_t i;
+  uint32_t current_nfthreads;
+
+  MUTEX_LOCK(q_mutex);
+  current_nfthreads = nfthreads;
+  max_nfthreads = new_count;
+  MUTEX_UNLOCK(q_mutex);
+
+  if (current_nfthreads > new_count) {
+    for (i = 0; i < current_nfthreads; i++) {
+      MUTEX_LOCK(q_mutex);
+      COND_SIGNAL(q_cond);
+      MUTEX_UNLOCK(q_mutex);
+    }
+  }
+}
+
 static void
 reset_ready_notify_pipe(void)
 {
@@ -511,29 +537,33 @@ send_ready_notify(void)
 static void
 create_pid_file(void)
 {
-#ifndef WIN32
   FILE *pid_file = NULL;
-  pid_t pid;
 
   if (!pid_file_path) {
     return;
   }
 
   pid_file = fopen(pid_file_path, "w");
-  pid = getpid();
-  fprintf(pid_file, "%d\n", pid);
+  {
+#ifdef WIN32
+    DWORD pid;
+    pid = GetCurrentProcessId();
+    fprintf(pid_file, "%" GRN_FMT_DWORD "\n", pid);
+#else /* WIN32 */
+    pid_t pid;
+    pid = getpid();
+    fprintf(pid_file, "%d\n", pid);
+#endif /* WIN32 */
+  }
   fclose(pid_file);
-#endif
 }
 
 static void
 clean_pid_file(void)
 {
-#ifndef WIN32
   if (pid_file_path) {
-    unlink(pid_file_path);
+    grn_unlink(pid_file_path);
   }
-#endif
 }
 
 static int
@@ -1003,7 +1033,7 @@ typedef struct {
 
 #define STRING_EQUAL_CI(string, string_length, constant_string)\
   (string_length == strlen(constant_string) &&\
-   strncasecmp(string, constant_string, string_length) == 0)
+   grn_strncasecmp(string, constant_string, string_length) == 0)
 
 static const char *
 do_htreq_post_parse_header_request_line(grn_ctx *ctx,
@@ -1903,10 +1933,13 @@ h_worker(void *arg)
   ht_context hc;
   grn_ctx ctx_, *ctx = &ctx_;
   grn_ctx_init(ctx, 0);
+  if (use_windows_event_log) {
+    grn_windows_event_logger_set(ctx);
+  }
   grn_ctx_use(ctx, (grn_obj *)arg);
   grn_ctx_recv_handler_set(ctx, h_output, &hc);
-  GRN_LOG(&grn_gctx, GRN_LOG_NOTICE, "thread start (%d/%d)", nfthreads, nthreads + 1);
   MUTEX_LOCK(q_mutex);
+  GRN_LOG(&grn_gctx, GRN_LOG_NOTICE, "thread start (%d/%d)", nfthreads, nthreads);
   do {
     grn_obj *msg;
     nfthreads++;
@@ -1916,6 +1949,10 @@ h_worker(void *arg)
         nfthreads--;
         goto exit;
       }
+      if (nthreads > max_nfthreads) {
+        nfthreads--;
+        goto exit;
+      }
     }
     nfthreads--;
     MUTEX_UNLOCK(q_mutex);
@@ -1927,8 +1964,8 @@ h_worker(void *arg)
   } while (nfthreads < max_nfthreads && grn_gctx.stat != GRN_CTX_QUIT);
 exit :
   nthreads--;
-  MUTEX_UNLOCK(q_mutex);
   GRN_LOG(&grn_gctx, GRN_LOG_NOTICE, "thread end (%d/%d)", nfthreads, nthreads);
+  MUTEX_UNLOCK(q_mutex);
   grn_ctx_fin(ctx);
   return GRN_THREAD_FUNC_RETURN_VALUE;
 }
@@ -1964,6 +2001,9 @@ h_server(char *path)
   int exit_code = EXIT_FAILURE;
   grn_ctx ctx_, *ctx = &ctx_;
   grn_ctx_init(ctx, 0);
+  if (use_windows_event_log) {
+    grn_windows_event_logger_set(ctx);
+  }
   MUTEX_INIT(q_mutex);
   COND_INIT(q_cond);
   CRITICAL_SECTION_INIT(cache_lock);
@@ -1978,8 +2018,8 @@ h_server(char *path)
 static grn_thread_func_result CALLBACK
 g_worker(void *arg)
 {
-  GRN_LOG(&grn_gctx, GRN_LOG_NOTICE, "thread start (%d/%d)", nfthreads, nthreads + 1);
   MUTEX_LOCK(q_mutex);
+  GRN_LOG(&grn_gctx, GRN_LOG_NOTICE, "thread start (%d/%d)", nfthreads, nthreads);
   do {
     grn_ctx *ctx;
     grn_edge *edge;
@@ -1990,6 +2030,10 @@ g_worker(void *arg)
         nfthreads--;
         goto exit;
       }
+      if (nthreads > max_nfthreads) {
+        nfthreads--;
+        goto exit;
+      }
     }
     ctx = &edge->ctx;
     nfthreads--;
@@ -2034,8 +2078,8 @@ g_worker(void *arg)
   } while (nfthreads < max_nfthreads && grn_gctx.stat != GRN_CTX_QUIT);
 exit :
   nthreads--;
-  MUTEX_UNLOCK(q_mutex);
   GRN_LOG(&grn_gctx, GRN_LOG_NOTICE, "thread end (%d/%d)", nfthreads, nthreads);
+  MUTEX_UNLOCK(q_mutex);
   return GRN_THREAD_FUNC_RETURN_VALUE;
 }
 
@@ -2099,6 +2143,9 @@ g_handler(grn_ctx *ctx, grn_obj *msg)
     edge = grn_edges_add(ctx, &((grn_msg *)msg)->edge_id, &added);
     if (added) {
       grn_ctx_init(&edge->ctx, 0);
+      if (use_windows_event_log) {
+        grn_windows_event_logger_set(&edge->ctx);
+      }
       GRN_COM_QUEUE_INIT(&edge->recv_new);
       GRN_COM_QUEUE_INIT(&edge->send_old);
       grn_ctx_use(&edge->ctx, (grn_obj *)com->ev->opaque);
@@ -2125,6 +2172,9 @@ g_server(char *path)
   int exit_code = EXIT_FAILURE;
   grn_ctx ctx_, *ctx = &ctx_;
   grn_ctx_init(ctx, 0);
+  if (use_windows_event_log) {
+    grn_windows_event_logger_set(ctx);
+  }
   MUTEX_INIT(q_mutex);
   COND_INIT(q_cond);
   CRITICAL_SECTION_INIT(cache_lock);
@@ -2150,6 +2200,7 @@ enum {
 #define FLAG_MODE_DAEMON     (1 << 6)
 #define FLAG_MODE_SERVER     (1 << 7)
 #define FLAG_NEW_DB     (1 << 8)
+#define FLAG_USE_WINDOWS_EVENT_LOG (1 << 9)
 
 static uint32_t
 get_core_number(void)
@@ -2370,7 +2421,7 @@ static const int default_http_port = DEFAULT_HTTP_PORT;
 static const int default_gqtp_port = DEFAULT_GQTP_PORT;
 static grn_encoding default_encoding = GRN_ENC_DEFAULT;
 static uint32_t default_max_num_threads = DEFAULT_MAX_NFTHREADS;
-static const int default_log_level = GRN_LOG_DEFAULT_LEVEL;
+static const grn_log_level default_log_level = GRN_LOG_DEFAULT_LEVEL;
 static const char * const default_protocol = "gqtp";
 static const char *default_hostname = "localhost";
 static const char * const default_dest = "localhost";
@@ -2598,7 +2649,10 @@ show_usage(FILE *output)
           "\n"
           "Logging options:\n"
           "  -l, --log-level <log level>:\n"
-          "                           specify log level (default: %d)\n"
+          "                           specify log level\n"
+          "                           [none|emergency|alert|critical|\n"
+          "                            error|warning|notice|info|debug|dump]\n"
+          "                           (default: %s)\n"
           "      --log-path <path>:   specify log path\n"
           "                           (default: %s)\n"
           "      --log-rotate-threshold-size <threshold>:\n"
@@ -2607,6 +2661,10 @@ show_usage(FILE *output)
           "                           log file size is larger than or\n"
           "                           equals to the threshold\n"
           "                           (default: 0; disabled)\n"
+#ifdef WIN32
+          "      --use-windows-event-log:\n"
+          "                           report logs as Windows events\n"
+#endif /* WIN32 */
           "      --query-log-path <path>:\n"
           "                           specify query log path\n"
           "                           (default: %s)\n"
@@ -2642,7 +2700,8 @@ show_usage(FILE *output)
           default_gqtp_port, default_bind_address,
           default_http_port, default_gqtp_port, default_hostname, default_protocol,
           default_document_root, default_cache_limit, default_max_num_threads,
-          default_log_level, default_log_path, default_query_log_path,
+          grn_log_level_to_string(default_log_level),
+          default_log_path, default_query_log_path,
           default_config_path, default_default_command_version,
           (long long int)default_default_match_escalation_threshold,
           default_dest);
@@ -2658,7 +2717,7 @@ main(int argc, char **argv)
   const char *bind_address_arg = NULL;
   const char *hostname_arg = NULL;
   const char *protocol_arg = NULL;
-  const char *log_path_arg = NULL;
+  const char *log_path_arg = GRN_LOG_PATH;
   const char *log_rotate_threshold_size_arg = NULL;
   const char *query_log_path_arg = NULL;
   const char *query_log_rotate_threshold_size_arg = NULL;
@@ -2704,6 +2763,8 @@ main(int argc, char **argv)
     {'\0', "input-fd", NULL, 0, GETOPT_OP_NONE},
     {'\0', "output-fd", NULL, 0, GETOPT_OP_NONE},
     {'\0', "working-directory", NULL, 0, GETOPT_OP_NONE},
+    {'\0', "use-windows-event-log", NULL,
+     FLAG_USE_WINDOWS_EVENT_LOG, GETOPT_OP_ON},
     {'\0', NULL, NULL, 0, 0}
   };
   opts[0].arg = &port_arg;
@@ -2876,6 +2937,16 @@ main(int argc, char **argv)
     do_server = g_server;
   }
 
+#ifdef WIN32
+  if (flags & FLAG_USE_WINDOWS_EVENT_LOG) {
+    use_windows_event_log = GRN_TRUE;
+  }
+#endif /* WIN32 */
+
+  if (use_windows_event_log) {
+    grn_windows_event_logger_set(NULL);
+  }
+
   if (log_path_arg) {
     grn_default_logger_set_path(log_path_arg);
   }
@@ -2913,19 +2984,29 @@ main(int argc, char **argv)
     grn_default_query_logger_set_rotate_threshold_size(value);
   }
 
-  if (log_level_arg) {
-    const char * const end = log_level_arg + strlen(log_level_arg);
-    const char *rest = NULL;
-    const int value = grn_atoi(log_level_arg, end, &rest);
-    if (end != rest || value < 0 || value > 9) {
-      fprintf(stderr, "invalid log level: <%s>\n", log_level_arg);
-      return EXIT_FAILURE;
+  {
+    grn_log_level log_level;
+
+    if (log_level_arg) {
+      grn_bool parsed;
+
+      parsed = grn_log_level_parse(log_level_arg, &log_level);
+      if (!parsed) {
+        const char * const end = log_level_arg + strlen(log_level_arg);
+        const char *rest = NULL;
+        const int value = grn_atoi(log_level_arg, end, &rest);
+        if (end != rest || value < GRN_LOG_NONE || value > GRN_LOG_DUMP) {
+          fprintf(stderr, "invalid log level: <%s>\n", log_level_arg);
+          return EXIT_FAILURE;
+        }
+        log_level = value;
+      }
+    } else {
+      log_level = default_log_level;
     }
-    log_level = value;
-  } else {
-    log_level = default_log_level;
+
+    grn_default_logger_set_max_level(log_level);
   }
-  grn_default_logger_set_max_level(log_level);
 
   if (max_num_threads_arg) {
     const char * const end = max_num_threads_arg + strlen(max_num_threads_arg);
@@ -2941,6 +3022,9 @@ main(int argc, char **argv)
     max_nfthreads = default_max_num_threads;
   }
 
+  grn_thread_set_get_count_func(groonga_get_thread_count, NULL);
+  grn_thread_set_set_count_func(groonga_set_thread_count, NULL);
+
   if (input_path) {
     if (!freopen(input_path, "r", stdin)) {
       fprintf(stderr, "can't open input file: %s (%s)\n",
